#!/usr/bin/perl

# author: Sergio Mover
#
# Creates a table from run logs of benchmarks.
#
# out is
#
# time mem res | time mem res | ...
# for every method tested.
#

use strict;

#------------------------------------------------------------------------------#
# parameters
#

# BASE
my $BASE;
my $hn = `hostname`;
$BASE = "..";

my @models_dir=("ROD",
                "fisher_star",
                "fisher_ring",
                "dist_controller",
                "bmw2",
                "audio");

my @meths = ("inter_step_sem_local_time",
             "inter_step_sem_global",
             "scenario_step_sem_invar",
             "scenario_step_sem_invar_avoid2timed",
             "bmc_hybrid_step_sem_local_time",
             "bmc_hybrid_step_sem_global");

#------------------------------------------------------------------------------#
# Format used for the column of each method.
# Data stored in the columns are:
#   - time
#   - memory
#   - results (re: reachable, un: un-reachable)
#   - length of the found path
#   - number of backtracks
#   - number of learned constraints
#   - search time without invar computation
#------------------------------------------------------------------------------#
my $exp_fmt = " %6s %6s %6s %6s %6s %6s %6s |";

my $reachable_res = "re";
my $unreachable_res = "un";
my $unknown_res = "na";


#------------------------------------------------------------------------------#
# global variables
#

#------------------------------------------------------------------------------#
# usage
#
sub usage() {
    printf "usage: generatetab.pl [logfile]\n";
    exit 1;
}


#------------------------------------------------------------------------------#
# print_exp
#
sub print_exp($$$$$$$$) {
    my $time = shift;
    my $mem = shift;
    my $res = shift;
    my $path_length = shift;
    my $backtrack_count = shift;
    my $learned_count = shift;
    my $info = shift;
    my $search_time_scen = shift;

# original printf
    printf $exp_fmt, $time, $mem, $search_time_scen, $path_length, $res,
    $backtrack_count, $learned_count;

    # test for bounded... do not try!
    # my $rounded = sprintf("%.3f", $time);
    # printf $exp_fmt, $rounded, $mem, $path_length, $res;
}

#------------------------------------------------------------------------------#
# evaluate_exp
#
sub evaluate_exp($$) {
    my $logfile = shift;
    my $meth =shift;

    my $time;
    my $mem;
    my $res;
    my $path_length;
    my $info;
    my $search_time_scen;

    my $backtrack_count;
    my $learned_count;

    ($time, $mem, $res, $path_length, $backtrack_count,
    $learned_count, $info, $search_time_scen) = evaluate_allmeths($logfile, $meth);

    return ($time, $mem, $res, $path_length, $backtrack_count,
            $learned_count, $info, $search_time_scen);
}

#------------------------------------------------------------------------------#
# evaluate_allmeths
#
sub evaluate_allmeths($$) {
    my $logfile = shift;
    my $meth = shift;

    my $line;
    my $time;
    my $mem;
    my $res;
    my $info;
    my $count;
    my $search_time_scen;

    my $elapse_count;
    my $path_length;
    my $backtrack_count; # total number of backtracks
    my $learned_count; # total number of learned constraints
    my $time_last_check; # time needed for the last checl

    open(LOG, $logfile) or die "error: can't open $logfile for reading: $!";
    $search_time_scen = $res = $time = $mem = $path_length = "na";
    $count = 0;

    if ($meth =~ /.*comp.*/) {
        $backtrack_count = $learned_count = 0;
    }
    else {
        # backtrack and learned constraints do make sense only for the
        # compositional approach.
        $backtrack_count = $learned_count = "na";
    }

    while (!($res eq "to" || $res eq "mo" || $res eq "er") && ($line = <LOG>)) {
        my $temp_res;
        my $found_res;

        my $temp_length;
        my $found_path;

        # Check the result of search
        if ($meth =~ /.*inter.*/) {
            ($found_res, $temp_res) = eval_res_invar($line);
        } elsif ($meth =~ /.*comp.*/) {
            ($found_res, $temp_res) = eval_res_comp($line);
        } elsif ($meth =~ /.*inc.*/) {
            ($found_res, $temp_res) = eval_res_incr($line);
        } elsif ($meth =~ /.*bmc\_hybrid.*/) {
            ($found_res, $temp_res) = eval_res_invar($line);
        } elsif ($meth =~ /.*scenario.*/) {
            ($found_res, $temp_res) = eval_res_scen($line);
        } elsif ($meth =~ /.*bach.*/) {
            ($found_res, $temp_res) = eval_res_bach($line);
        } else {
            die "error: can't match method ${meth}\n"
        }


        if ($found_res == 1) {
            # if ($logfile =~ "../dist_controller_m5/dist_controller_4_2.scenario_step_sem.txt") {
            #     printf "\n$logfile $res $temp_res \n";
            # }

            if ($res =~ "re") {
                $res = "re";
            }
            else {
                $res = $temp_res;
            }
        }

        # Check for path length
        if ($meth =~ /.*inter.*/) {
            ($found_path, $temp_length) = get_path_length_invar($line);
        } elsif ($meth =~ /.*bmc\_hybrid.*/) {
            ($found_path, $temp_length) = get_path_length_hybrid($line);
        } elsif ($meth =~ /.*comp.*/) {
            ($found_path, $temp_length) = get_path_length_comp($line);
        } elsif ($meth =~ /.*inc.*/) {
            ($found_path, $temp_length) = get_path_length_incr($line);
        } elsif ($meth =~ /.*scenario.*/) {
            ($found_path, $temp_length) = get_path_length_scen($line);
        } elsif ($meth =~ /.*bach.*/) {
            # No path length for bach
            $found_path = 0;
        } else {
            die "error: can't match method ${meth}\n"
        }

        if ($found_path == 1) {
            $path_length = $temp_length;
        }

        # check for backtrack count and learned constraints count
        if ($meth =~ /.*comp*./) {
            ($backtrack_count) = get_backtracks_count_comp($line, $backtrack_count);
            ($learned_count) = get_learned_count_comp($line, $learned_count);
        }

        # get the search time
        if ($meth =~ /.*scenario.*/) {
            if ($line =~ /SEARCH TIME:\s+([0-9.]+).*/) {
                $search_time_scen = $1;
                $search_time_scen = substr ($search_time_scen,0, 5);
            }
        }


        if($line =~ /^.*elapse:\s*(\d+\.\d+).*$/) {
            # get the time between two time commands (time elapsed for the check!)
            $count ++;

            if ($count == 2) {
                $time = $1;
            }
        } elsif ($line =~ /\[run\] space:\s+([\-0-9\.]+) MB/) {
            $mem = $1;
        } elsif ($line =~ /\[run\] status:\s+out of time/) {
            $res = $time = $mem = "to";
            $path_length = "na";
        } elsif ($line =~ /\[run\] status:\s+out of memory/) {
            $res = $time = $mem = "mo";
            $path_length = "na"
        } elsif ($line =~ /\[run\] status:\s+segmentation fault/) {
            $res = "er";
            $time = $mem = "to";
            $path_length = "na";
        } elsif ($line =~ /\[run\] status:\s+signal\(6\)/) {
            $res = "er";
            $time = $mem = "to";
            $path_length = "na";
        } elsif ($line =~ /\[run\] status:\s+signal\(9\)/) {
            $res = "er";
            $time = $mem = "to";
            $path_length = "na";
        } elsif ($line =~ /Killed/) {
            $res = "er";
            $time = $mem = "to";
            $path_length = "na";
        } elsif ($line =~ /result = NULL/) {
            $res = "er";
            $time = $mem = "to";
            $path_length = "na";
        } elsif ($line =~ /input buffer overflow/) {
            $res = "er";
            $time = $mem = "to";
            $path_length = "na";
        } elsif ($meth =~ /.*bach.*/) {
            # find time result for bach
            if ($line =~ /^([0-9]+)$/) {
                my $time_sec=$1/1000;
                $time = $time_sec;
            }
        }
    }
    close(LOG);


    if ($res eq $unreachable_res || $res eq "er" || $res eq $unknown_res) {
        $time = "to";
        $mem = "to";
        $path_length = "na";
        $search_time_scen = "na";
    }

    return ($time, $mem, $res, $path_length, $backtrack_count,
            $learned_count, $info, $search_time_scen);
}


#------------------------------------------------------------------------------#
# eval_res_invar
# Evaluate the results of an invariant checking
#
sub eval_res_invar($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

# match results.
# For invariant: "-- invariant TRUE is true" or
# "invariant FALSE is false, a counter-example of size 0 exists"
# -- no counterexample found for invariant t1.state = Far up to 0

    if ($line =~ /-- invariant .* is true.*/) {
        # return unreachable!
        $res = "${unreachable_res}";
        $found_res = 1;
    } elsif ($line =~ /invariant .* is false.*/) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    } elsif ($line =~ /The property is false, a counter-example of size [0-9]+ exists.*/) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    } elsif ($line =~ /-- no counterexample found for invariant.*/) {
        # return unknown!
        $res = "${unknown_res}";
        $found_res = 1;
    }

    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# eval_res_scen
# Evaluate the results of an scenario checking
#
sub eval_res_scen($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

# match results.
# -- The scenario is not satisfiable
# -- The scenario is reachable with k local steps
# No counterexample found for the scenario up to

    if ($line =~ /-- The scenario is not satisfiable.*/) {
        # return unreachable!
        $res = "${unreachable_res}";
        $found_res = 1;
    } elsif ($line =~ /-- The scenario is reachable with k local steps.*/) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    } elsif ($line =~ /-- as demonstrated by the following execution sequence.*/) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    } elsif ($line =~ /No counterexample found for the scenario up to.*/) {
        # return unknown!
        $res = "${unknown_res}";
        $found_res = 1;
    }

    return ($found_res, $res);
}




#------------------------------------------------------------------------------#
# eval_res_incr
# Evaluate the results of an incremental checking
#
sub eval_res_incr($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

# match results.
# "The target is reachable."
# "The target was not reached but the search is not complete."

    if ($line =~ /The target was not reached but the search is not complete\./) {
        # return unknown!
        $res = "${unknown_res}";
        $found_res = 1;
    } elsif ($line =~ /The target is reachable\./) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    }
    # elsif ($line =~ /The target was not reached but the search is not complete\./) {
    #      # return unreachable!
    #     $res = "${unreachable_res}";
    #     $found_res = 1;
    # }
    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# eval_res_comp
# Evaluate the results of compositional checking
#
sub eval_res_comp($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

# match results.
# For invariant: "-- invariant TRUE is true" or
# "invariant FALSE is false, a counter-example of size 0 exists"
# -- no counterexample found for invariant t1.state = Far up to 0

    if ($line =~ /-- invariant .* is true.*/) {
        # return unreachable!
        $res = "${unreachable_res}";
        $found_res = 1;
    } elsif ($line =~ /-- as demonstrated by the following execution sequence*/) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    } elsif ($line =~ /-- no counterexample found for invariant.*/) {
        # return unknown!
        $res = "${unknown_res}";
        $found_res = 1;
    }

    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# eval_res_bach
# Evaluate the results of bach
#
sub eval_res_bach($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

    # if ($line =~ //) {
    #     # return unknown!
    #     $res = "${unknown_res}";
    #     $found_res = 1;
    if ($line =~ /the target can be satisfied/) {
        # return reachable!
        $res = "${reachable_res}";
        $found_res = 1;
    }
    # elsif ($line =~ /The target was not reached but the search is not complete\./) {
    #      # return unreachable!
    #     $res = "${unreachable_res}";
    #     $found_res = 1;
    # }
    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# get the path length found by an invariant checking
#
sub get_path_length_invar($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

    if ($line =~ /is false, a counter-example of size ([0-9]+) exists/) {
        $res = $1;
        $found_res = 1;
    }

    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# get the path length found by an bmc for hybrid checking
#
sub get_path_length_hybrid($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;


    if ($line =~ /-> State: [0-9].([0-9]+) <-/) {
        $res = $1;
        $found_res = 1;
    }

    return ($found_res, $res);
}


#------------------------------------------------------------------------------#
# get the path length found by compositional checking
#
sub get_path_length_comp($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

    if ($line =~ /is reachable at step ([0-9]+)/) {
        $res = $1;
        $found_res = 1;
    }

    return ($found_res, $res);
}



#------------------------------------------------------------------------------#
# get the path length found by an incremental checking
#
sub get_path_length_incr($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

    if ($line =~ / -- path reaching the target found in ([0-9]+) steps/) {
        $res = $1;
        $found_res = 1;
    }

    if ($line =~ / -- path reaching the targetfound in ([0-9]+) steps/) {
        $res = $1;
        $found_res = 1;
    }

    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# get the path length found for the scenario checking
#
sub get_path_length_scen($)
{
    my $line = shift;
    my $found_res = 0;
    my $res;

    if ($line =~ /-- The scenario is reachable with k local steps ([0-9]+)/) {
        $res = $1;
        $found_res = 1;
    }

    return ($found_res, $res);
}

#------------------------------------------------------------------------------#
# increase the number of backtracks
#
sub get_backtracks_count_comp($$)
{
    my $line = shift;
    my $backtrack_count = shift;

    if ($line =~ /PROCESS UNSAT: ID: ([0-9]+)/) {
        $backtrack_count = $backtrack_count + 1;
    }

    return $backtrack_count;
}

#------------------------------------------------------------------------------#
# increase the number of learned constraints
#
sub get_learned_count_comp($$)
{
    my $line = shift;
    my $learned_count = shift;

    if ($line =~ /NUMBER OF LEARNED CLAUSES: ([0-9]+)/) {
        $learned_count = $learned_count + $1;
    }

    return $learned_count;
}

#------------------------------------------------------------------------------#
# main
#
{
    my $logfile;

#------------------------------------------------------------------------------#
# read parameters
#
    {
        my $argno=0;

        while ($ARGV[$argno] ne "") {
            if ($ARGV[$argno] eq "-h" || $ARGV[$argno] eq "--help") {
                $argno++;
                usage;
            } else {
                $logfile=$ARGV[$argno];
                $argno++;
            }
        }
    }

#------------------------------------------------------------------------------#
# do work
#
    {
        my $time;
        my $constructiontime;
        my $searchtime;
        my $mem;
        my $res;
        my $path_length;
        my $backtrack_count;
        my $learned_count;
        my $info;
        my $search_time_scen;

        if ($logfile eq "") {
            my $dir;
            my $meth;
            my @lines;

            foreach $dir (@models_dir) {
                my $exists = "f";
                my $dir_long = sprintf "$BASE/${dir}/";
                my %model_hash;
                my $model_file_name;
                my $model_name;

# fihser ring, star
# find . -name "*hydi" | sort -t '_' -n -k4 -k3`
# ETCS hysat
# find . -name "*hydi" | sort -t '_' -n -k5 -k4
# find FDDI
# find . -name "*hydi" | sort -t '_' -n -k3 -k2

                if ($dir =~ /fisher_ring/ ||
                    $dir =~ /fisher_star/ ||
                    $dir =~ /motorcycle_invariant*/ ||
                    $dir =~ /FDDI_invariant*/ ||
                    $dir =~ /dist_controller/) {
                    @lines = `ls ${dir_long}*.scen | sort -g -t '_' -n -k4 -k3`;
                } elsif ($dir =~ /ETCS_hysat*/ ||
                         $dir =~ /fisher_ring_*/ ||
                         $dir =~ /fisher_ring_*/) {
                    @lines =  `ls ${dir_long}*.scen | sort -t '_' -n -k5 -k4`;
                } elsif ($dir =~ /FDDI/) {
                    @lines = `ls ${dir_long}*.scen | sort -t '_' -n -k3 -k2`;
                } elsif ($dir =~ /bmw2/) {
                    @lines = `ls ${dir_long}*.scen | sort -t '_' -n -k2`;
                } elsif ($dir =~ /bmw/) {
                    @lines = `ls ${dir_long}*.scen | sort -t '_' -n -k2`;
                } else {
                    # Get the models contained in model dir
                    @lines = `ls -1 ${dir_long}*.scen`;
                }

                my $model_prefix;
                foreach $model_file_name (@lines) {
                    my $model_name;


                    if ($model_file_name =~ /ETCS.*/) {
                        # workaround for etcs models
                        # $model_file_name =~ /${dir_long}ETCS_TrainsRBC6_t([0-9]+)_s3\.scen/;
                        # $model_name = "ETCS_TrainsRBC6_t$1_s3";
                        # $model_prefix = "${dir_long}ETCS_TrainsRBC6_t${model_name}_s3";
                        $model_file_name =~ /${dir_long}ETCS_TrainsRBC6_([0-9]+)\.scen/;
                        $model_name = "ETCS_TrainsRBC6_$1_3";
                        $model_prefix = "${dir_long}ETCS_TrainsRBC6_${model_name}";
                   }
                    # if ($model_file_name =~ /bmw.*/) {
                    #     # workaround for bmw models
                    #     $model_file_name =~ /${dir_long}bmw([0-9]*)_lin_n([0-9]+)l1.scen/;
                    #     $model_name = "bmw$1_lin_n$2l1";
                    #     $model_prefix = "${dir_long}bmw$1_lin_n$2l1";
                    # if ($model_file_name =~ /bmw.*/) {
                    #     # workaround for bmw models
                    #     $model_file_name =~ /${dir_long}bmwlin_([0-9]*).scen/;
                    #     $model_name = "bmwlin_20_5_$1";
                    #     $model_prefix = "${dir_long}bmwlin_20_5_$1";
                    my $appbmw;
                    if($dir =~ /bmw/) {
                        $appbmw="1";
                    }
                    if ($dir =~ /bmw2/) {
                        $appbmw="2";
                    }
                    if ($model_file_name =~ /bmwlin_([0-9]*).*/) {
                        # workaround for bmw models
                        $model_file_name =~ /${dir_long}bmwlin_${appbmw}_$1.scen/;
#                        $model_file_name =~ /${dir_long}([a-z\-_0-9]+)\.scen/;
                        $model_name="bmwlin_${appbmw}_$1";
                        $model_prefix = "${dir_long}bmwlin_${appbmw}";

                    }elsif ($model_file_name =~ /ETCS_hysat_noma.*/) {
                        $model_file_name =~ /${dir_long}ETCS_hysat_noma_([a-z\-_0-9]+).scen/;
                        $model_name = "ETCS_hysat_noma_$1";
                        $model_prefix = "${dir_long}ETCS_hysat_noma_${model_name}";
                    } else {
                        $model_file_name =~ /${dir_long}([a-z\-_0-9]+)\.scen/;
                        $model_name = $1;
                        $model_prefix = "${dir_long}${model_name}";
                    }

                    if (! (exists $model_hash{"$model_name"})) {
#                        print "\n ${model_name}\n";

                        $model_hash{"${model_name}"} = "1";
                    }
                }


                # ORDER MODEL NAME
                my @model_names_not_ordered=keys %model_hash;
                my $model_names_list=`echo @{model_names_not_ordered}`;

#                print "\n${model_names_list}\n";

                $model_names_list =~ s/ /\n/g;

                my @model_names_ordered;

                if ($dir =~ /fisher_ring*/ ||
                    $dir =~ /fisher_star*/) {
                    @model_names_ordered = `echo "${model_names_list}" | sort -t '_' -n -k4 -k3`;
                } elsif ($dir =~ /ETCS_hysat*/) {
                    @model_names_ordered =  `echo "${model_names_list}" | sort -t '_' -n -k5 -k4`;
                } elsif ($dir =~ /FDDI_invariant_more_bench*/) {
                    @model_names_ordered =  `echo "${model_names_list}" | sort -t '_' -n -k3 -k2`;
                } elsif ($dir =~ /FDDI*/ ||
                    $dir =~ /motorcycle*/) {
                    @model_names_ordered = `echo "${model_names_list}" | sort -t '_' -n -k3 -k2`;
                } elsif ($dir =~ /bmw/ ||
                    $dir =~ /bmw*/) {
                    @model_names_ordered = `echo "${model_names_list}" | sort -t '_' -n -k 1`;

                } else {
                    @model_names_ordered = `echo "${model_names_list}"`;
                }

                print "MODEL_DIR: ${dir}\n";
                my $max_length=25;
#                foreach $model_name (sort(keys %model_hash)) {
                foreach $model_name (@model_names_ordered) {
                    $model_name =~ s/\n//g;

                    if (! (${model_name} eq "")) {
                        # remove scenario from model name
                        $model_name =~ s/_scenario//;

                       print "${model_name}";


                        for (my $count = length $model_name; $count < $max_length; $count++) {
                            print " ";
                        }

                        if (length $model_name < $max_length) {
                        }
                        print "\t";

                        foreach $meth (@meths) {
                            my $exists = "f";

                            $logfile = "${dir_long}${model_name}.${meth}.txt";

#                            print "\nSearching ${logfile}\n";

                            $exists = (-f "${logfile}") ? "t" : $exists;

                            if ("${exists}" eq "t") {
                                if (-e $logfile) {
                                    ($time, $mem, $res, $path_length,
                                     $backtrack_count, $learned_count, $info, $search_time_scen) = evaluate_exp($logfile, $meth);

                                    print_exp($time, $mem, $res,
                                              $path_length, $backtrack_count,
                                              $learned_count, $info, $search_time_scen);
                                }
                            } else {
                                # print standarad value (na)
                                print_exp("to", "mo",
                                              "${unknown_res}", "na",
                                              "", "na", "na", "na");
#                            print_exp("na", "mo", "${unknown_res}", "na", "");
                            }
                        } # end of loop on methods

                        print "\n" # separates results lines
                    } # end of loop on found models
                } # end of if model name is empty
            }
        }
    }

#------------------------------------------------------------------------------#
# clean up and exit
#
    exit 0;
}
